/**
 * EasyBeans
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: WarArchiveHelper.java 3005 2008-04-29 10:31:45Z gaellalire $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.ee.metadata.war.impl;

import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.ow2.util.archive.api.ArchiveException;
import org.ow2.util.archive.api.IArchive;
import org.ow2.util.archive.impl.ArchiveManager;
import org.ow2.util.archive.impl.MemoryArchive;
import org.ow2.util.scan.api.DefaultArchiveHelper;
import org.ow2.util.scan.api.IArchiveHelper;

/**
 * This class is a temporary solution to solve war archive scan issue. Indeed an
 * instance of a class which implements ArchiveHelper should not be archive
 * dependent.
 * @author Gael Lalire
 */
public class WarArchiveHelper implements IArchiveHelper {

    /**
     * Class prefix.
     */
    private static final String CLASS_PREFIX = "WEB-INF/classes/";

    /**
     * Class extension.
     */
    private static final String CLASS_SUFFIX = ".class";

    /**
     * Lib prefix.
     */
    private static final String LIB_PREFIX = "WEB-INF/lib/";

    /**
     * Lib extension.
     */
    private static final String LIB_SUFFIX = ".jar";

    /**
     * The list of class inside war archive.
     */
    private List<URL> classURLList;

    /**
     * The list of lib inside war archive.
     */
    private List<IArchiveHelper> archiveHelperList;


    /**
     * The archive.
     */
    private IArchive archive;

    /**
     * ClassLoader.
     */
    // private URLClassLoader classLoader;


    /**
     * WarArchiveHelper constructor.
     * @param archive the archive
     * @throws ArchiveException if error occurs
     */
    public WarArchiveHelper(final IArchive archive) throws ArchiveException {
        throw new UnsupportedOperationException("please use WarArchiveHelper with classname contructor (temporary solution)");
    }


    /**
     * WarArchiveHelper constructor.
     * @param archive the archive
     * @param classNames the class names
     * @throws ArchiveException if error occurs
     */
    public WarArchiveHelper(final IArchive archive, final List<String> classNames) throws ArchiveException {
        this.classURLList = new ArrayList<URL>();
        this.archive = archive;
        archiveHelperList = new ArrayList<IArchiveHelper>();
        if (archive instanceof MemoryArchive) {
            findClassesWithEntriesScan((MemoryArchive) archive);
        } else {
            findClassesCreateClassLoader(classNames);
        }
    }

    /**
     * Gets URLs from the given WEB Archive.
     * @param archive a memory archive
     * @throws ArchiveException if the URLs are not retrieved
     */
    private void findClassesWithEntriesScan(final MemoryArchive archive) throws ArchiveException {
        Iterator<String> stringIterator = archive.getEntries();
        while (stringIterator.hasNext()) {
            String resourceName = stringIterator.next();
            if (resourceName.startsWith(CLASS_PREFIX) && resourceName.endsWith(CLASS_SUFFIX)) {
                classURLList.add(archive.getResource(resourceName));
            } else {
                if (resourceName.startsWith(LIB_PREFIX) && resourceName.endsWith(LIB_SUFFIX)) {
                    IArchive libArchive = ArchiveManager.getInstance().getArchive(archive.getResource(resourceName));
                    IArchiveHelper archiveHelper = new DefaultArchiveHelper(libArchive);
                    classURLList.addAll(archiveHelper.listClassURL());
                    archiveHelperList.add(archiveHelper);
                }
            }
        }
    }


    /**
     * Gets URLs from the given WEB Archive that are required for a classloader.
     * Should be deprecated when IArchive will have getEntries method.
     * @param classNames the class names
     * @throws ArchiveException if the URLs are not retrieved
     */
    private void findClassesCreateClassLoader(final List<String> classNames) throws ArchiveException {
        // Classloader of a war should have access to WEB-INF/classes and all jars of WEB-INF/lib
        // Build list of matching URLs.
        Iterator<URL> itResources;
        try {
            itResources = archive.getResources();
        } catch (ArchiveException e) {
            throw new ArchiveException("Cannot get resources on the given archive '" + archive + "'.", e);
        }
        List<URL> classLoaderURLs = new ArrayList<URL>();

        // Root URL of the WAR
        URL warURL = archive.getURL();

        // WEB-INF/classes ?
        /*
        URL webInfClassesURL = webArchive.getResource("WEB-INF/classes");
        boolean webinfClassesFound = false;
        if (webInfClassesURL != null) {
            classLoaderURLs.add(webInfClassesURL);
            webinfClassesFound = true;
        }
        */

        // For each resource, try to search matching WEB-INF/ URLs
        while (itResources.hasNext()) {
            URL currentURL = itResources.next();

            // add .class ending url to url list (should be in WEB-INF/classes)
            if (currentURL.getPath().endsWith(".class")) {
                classURLList.add(currentURL);
            }

            // Remove start of the URL
            int filePos = currentURL.toExternalForm().indexOf(warURL.toExternalForm());
            if (filePos >= 0) {
                filePos += warURL.toExternalForm().length();
            }
            // Compute relative path
            String relativePath = currentURL.toExternalForm().substring(filePos);
            if (relativePath.startsWith("!/")) {
                relativePath = relativePath.substring(2);
            } else if (relativePath.startsWith("/")) {
                relativePath = relativePath.substring(1);
            }

            // Check for WEB-INF/classes URL if not already found
            /*
            if (!webinfClassesFound) {
                if (relativePath.startsWith("WEB-INF/classes/")) {
                    int index = currentURL.toExternalForm().indexOf(relativePath);
                    URL webinfClassesURL = null;
                    try {
                        webinfClassesURL = new URL(currentURL.toExternalForm().substring(0, index) + "WEB-INF/classes/");
                    } catch (MalformedURLException e) {
                        throw new ArchiveException("Unable to build URL for WEB-INF/classes on the given archive '" + webArchive
                                + "'.", e);
                    }
                    classLoaderURLs.add(webinfClassesURL);
                    webinfClassesFound = true;
                }
            }
            */

            // Keep valid URLs
            if (relativePath.startsWith("WEB-INF/lib/") && relativePath.endsWith(".jar")) {
                classLoaderURLs.add(currentURL);
            }
        }

        // Temporary solution
        /*
        classLoader = new URLClassLoader(classLoaderURLs.toArray(new URL[classLoaderURLs.size()]));
        for (String className : classNames) {
            String classLocation = className.replace('.', '/').concat(".class");
            URL url = archive.getResource("WEB-INF/classes/" + classLocation);
            if (url == null) {
                classURLList.add(classLoader.getResource(classLocation));
            }
        }
        */

        for (URL url : classLoaderURLs) {
            IArchive archive = ArchiveManager.getInstance().getArchive(url);
            IArchiveHelper archiveHelper = new DefaultArchiveHelper(archive);
            classURLList.addAll(archiveHelper.listClassURL());
            archiveHelperList.add(archiveHelper);
        }

        // TODO transform URL in Archive then add to construct DefaultArchiveHelper and find all classes

    }


    /**
     * @param className a class name
     * @return url of className inside the archive
     * @throws ArchiveException if error occurs
     */
    public URL getClassURL(final String className) throws ArchiveException {
        String classLocation = className.replace('.', '/').concat(".class");
        URL url = archive.getResource("WEB-INF/classes/" + classLocation);
        if (url == null) {
            // search in WEB-INF/lib libs
            for (IArchiveHelper archiveHelper : archiveHelperList) {
                url = archiveHelper.getClassURL(className);
                if (url != null) {
                    break;
                }
            }
            // TODO search in ear lib

            // TODO search in classloader lib
        }
        return url;
    }

    /**
     * @return the list of all classes in archive
     */
    public List<URL> listClassURL() {
        return this.classURLList;
    }

    /**
     * @return the archive
     */
    public IArchive getArchive() {
        return archive;
    }

}
