/**
 * OW2 Util
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: WarDeployableMetadataFactory.java 2999 2008-04-28 16:14:09Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.ee.metadata.war.impl;

import java.util.LinkedList;
import java.util.List;

import org.ow2.util.archive.api.ArchiveException;
import org.ow2.util.archive.api.IArchive;
import org.ow2.util.ee.deploy.api.deployable.WARDeployable;
import org.ow2.util.ee.deploy.api.deployable.metadata.DeployableMetadataException;
import org.ow2.util.ee.metadata.war.api.IWarDeployableMetadata;
import org.ow2.util.ee.metadata.war.api.IWarDeployableMetadataFactory;
import org.ow2.util.ee.metadata.war.api.exceptions.WARDeploymentDescException;
import org.ow2.util.ee.metadata.war.impl.configurator.WarDeployableMetadataConfigurator;
import org.ow2.util.ee.metadata.war.impl.xml.Filter;
import org.ow2.util.ee.metadata.war.impl.xml.Listener;
import org.ow2.util.ee.metadata.war.impl.xml.Servlet;
import org.ow2.util.ee.metadata.war.impl.xml.Tag;
import org.ow2.util.ee.metadata.war.impl.xml.WAR;
import org.ow2.util.ee.metadata.war.impl.xml.WARDeploymentDesc;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.scan.api.ArchiveScanner;
import org.ow2.util.scan.api.ArchiveScannerFactory;
import org.ow2.util.scan.api.IArchiveScannerFactory;
import org.ow2.util.scan.api.ScanException;

/**
 * Factory to contruct war deployable metadata.
 * @author Gael Lalire
 */
public class WarDeployableMetadataFactory implements IWarDeployableMetadataFactory {

    /**
     * Logger.
     */
    private static final Log LOGGER = LogFactory.getLog(WarDeployableMetadataFactory.class);

    /**
     * Archive scanner factory.
     */
    private IArchiveScannerFactory archiveScannerFactory;

    /**
     * Constructor.
     * @throws ScanException if no archive scanner factory can be found
     */
    public WarDeployableMetadataFactory() throws ScanException {
        this(ArchiveScannerFactory.getInstance());
    }

    /**
     * Constructor with an archive scanner factory.
     * @param archiveScannerFactory the archive scanner factory
     */
    public WarDeployableMetadataFactory(final IArchiveScannerFactory archiveScannerFactory) {
        this.archiveScannerFactory = archiveScannerFactory;
    }

    /**
     * Create war deployable metadata.
     * @param deployable the war deployable
     * @return the metadata
     * @throws DeployableMetadataException if error occurs
     */
    public IWarDeployableMetadata createDeployableMetadata(final WARDeployable deployable) throws DeployableMetadataException {

        IArchive archive = deployable.getArchive();

        // Get WAR DD.
        WAR war;
        try {
            war = WARDeploymentDesc.getWAR(archive);
        } catch (WARDeploymentDescException e) {
            throw new DeployableMetadataException(e);
        }

        // for time debugging
        long tAnalyzeStart = System.currentTimeMillis();

        // Build list of class to analyze
        List<String> classNames = new LinkedList<String>();

        // Add the servlets
        for (Servlet servlet : war.getServlets()) {
            addClassIfNotAlreadyPresent(servlet.getClassName(), classNames);
        }

        // Add the listeners
        for (Listener listener : war.getListeners()) {
            addClassIfNotAlreadyPresent(listener.getListenerClassName(), classNames);
        }

        // Add the tags
        for (Tag tag : war.getTags()) {
            addClassIfNotAlreadyPresent(tag.getTagClassName(), classNames);
        }

        // Add the filters
        for (Filter filter : war.getFilters()) {
            addClassIfNotAlreadyPresent(filter.getFilterClassName(), classNames);
        }

        // TODO: Analyze super classes

        // Analyze the set of given classes

        ArchiveScanner archiveScanner;
        try {
            archiveScanner = this.archiveScannerFactory.createArchiveScanner(new WarArchiveHelper(archive, classNames));
        } catch (ArchiveException e) {
            throw new DeployableMetadataException(e);
        }

        WarDeployableMetadataConfigurator warArchiveMetadataConfigurator = new WarDeployableMetadataConfigurator(deployable);


        archiveScanner.addArchiveConfigurator(warArchiveMetadataConfigurator);

        try {
            archiveScanner.scanCollection(classNames);
        } catch (ScanException e) {
            throw new DeployableMetadataException(e);
        }
        archiveScanner.endScan();

        // time if debugging
        if (LOGGER.isDebugEnabled()) {
            long tAnalyzeStartEnd = System.currentTimeMillis();
            if (LOGGER.isDebugEnabled()) {
                LOGGER.debug("Analyze of file {0} took {1} ms.", archive.getName(), (tAnalyzeStartEnd - tAnalyzeStart));
            }
        }

        return warArchiveMetadataConfigurator.getWarDeployableMetadata();
    }

    /**
     * Add a given class in the list only if it is not yet present in the list.
     * @param className the given name of the class
     * @param classNames the lsit of existing classes.
     */
    protected void addClassIfNotAlreadyPresent(final String className, final List<String> classNames) {
        // Sometimes, servlet classname is null as it is referenced as a
        // jsp-file
        if (className == null) {
            return;
        }
        if (!classNames.contains(className)) {
            classNames.add(className);
        }
    }

    /**
     * @return the deployable metadata class
     */
    public Class<IWarDeployableMetadata> getDeployableMetadataClass() {
        return IWarDeployableMetadata.class;
    }

}
