/**
 * EasyBeans
 * Copyright (C) 2007-2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: ENCBindingBuilder.java 2978 2008-04-27 15:42:41Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.ee.metadata.war.impl.enc;

import java.util.Arrays;
import java.util.List;

import org.ow2.easybeans.asm.Type;
import org.ow2.util.ee.metadata.common.api.ICommonClassMetadata;
import org.ow2.util.ee.metadata.common.api.ICommonFieldMetadata;
import org.ow2.util.ee.metadata.common.api.ICommonMethodMetadata;
import org.ow2.util.ee.metadata.common.api.interfaces.ISharedMetadata;
import org.ow2.util.ee.metadata.common.api.struct.IJAnnotationResource;
import org.ow2.util.ee.metadata.common.api.struct.IJEjbEJB;
import org.ow2.util.ee.metadata.common.api.struct.IJavaxPersistenceContext;
import org.ow2.util.ee.metadata.common.api.struct.IJavaxPersistenceUnit;
import org.ow2.util.ee.metadata.common.impl.CommonFieldMetadata;
import org.ow2.util.ee.metadata.common.impl.CommonMethodMetadata;
import org.ow2.util.ee.metadata.war.api.IWarDeployableMetadata;
import org.ow2.util.scan.api.metadata.structures.JMethod;

/**
 * Utility class that helps to build the ENC Binding by analyzing metadata.
 * @author Florent BENOIT
 */
public class ENCBindingBuilder {

    /**
     * All class metadata are kept by this object.
     */
    private IWarDeployableMetadata warArchiveMetadata = null;

    /**
     * Binding holder object.
     */
    private ENCBindingHolder bindingHolder = null;

    /**
     * Replace length to create default JNDI names.
     */
    private static final int GETTER_LENGTH = 3;

    /**
     * Constructor used only by this class/subclasses.
     * @param warArchiveMetadata the given metadata to analyze.
     */
    protected ENCBindingBuilder(final IWarDeployableMetadata warArchiveMetadata) {
        this.warArchiveMetadata = warArchiveMetadata;
        this.bindingHolder = new ENCBindingHolder();
    }

    /**
     * Analyze the metadata and populate the binding holder.
     * @throws ENCBindingException if the analyze is failing.
     */
    protected void analyze() throws ENCBindingException {

        // Analyze each class
        for (ICommonClassMetadata<?, ?, ?> classMetadata : this.warArchiveMetadata.getWarClassMetadataCollection()) {
            analyzeClassMetadata(classMetadata);
        }
    }

    /**
     * Analyze the given class metadata.
     * @param classMetadata the given metadata to analyze.
     * @throws ENCBindingException if the class has not been analyzed
     */
    protected void analyzeClassMetadata(final ICommonClassMetadata<?, ?, ?> classMetadata) throws ENCBindingException {
        // Are there persistence contexts (only on Class)
        List<IJavaxPersistenceContext> javaxPersistenceContexts = classMetadata.getJavaxPersistencePersistenceContexts();
        if (javaxPersistenceContexts != null) {
            for (IJavaxPersistenceContext javaxPersistenceContext : javaxPersistenceContexts) {
                addPersistenceContext(javaxPersistenceContext);
            }
        }

        // Are there persistence units (only on Class)
        List<IJavaxPersistenceUnit> javaxPersistenceUnits = classMetadata.getJavaxPersistencePersistenceUnits();
        if (javaxPersistenceUnits != null) {
            for (IJavaxPersistenceUnit javaxPersistenceUnit : javaxPersistenceUnits) {
                addPersistenceUnit(javaxPersistenceUnit);
            }
        }

        // Are there resources (only on Class)
        List<IJAnnotationResource> jAnnotationResources = classMetadata.getJAnnotationResources();
        if (jAnnotationResources != null) {
            for (IJAnnotationResource jAnnotationResource : jAnnotationResources) {
                addResource(jAnnotationResource);
            }
        }

        // Are there EJBs (only on Class)
        List<IJEjbEJB> jEjbEJBs = classMetadata.getJEjbEJBs();
        if (jEjbEJBs != null) {
            for (IJEjbEJB jEjbEJB : jEjbEJBs) {
                addEJB(jEjbEJB);
            }
        }

        // Analyze common stuff on the class
        analyzeCommonMetadata(classMetadata);

        // Then, analyze each field of the class
        for (ICommonFieldMetadata<?, ?, ?> fieldAnnotationMetadata : classMetadata.getFieldMetadataCollection()) {
            analyzeFieldMetadata(fieldAnnotationMetadata);
        }

        // And each method of the class
        for (ICommonMethodMetadata<?, ?, ?> methodAnnotationMetadata : classMetadata.getMethodMetadataCollection()) {
            analyzeMethodMetadata(methodAnnotationMetadata);
        }

    }

    /**
     * Analyze the given field metadata.
     * @param fieldMetadata the given metadata to analyze.
     * @throws ENCBindingException if the field has not been analyzed
     */
    protected void analyzeFieldMetadata(final ICommonFieldMetadata<?, ?, ?> fieldMetadata) throws ENCBindingException {
        // Analyze common stuff on the field
        analyzeCommonMetadata(fieldMetadata);
    }

    /**
     * Analyze the given method metadata.
     * @param methodMetadata the given metadata to analyze.
     * @throws ENCBindingException if the method has not been analyzed
     */
    protected void analyzeMethodMetadata(final ICommonMethodMetadata<?, ?, ?> methodMetadata) throws ENCBindingException {
        // Analyze common stuff on the method
        analyzeCommonMetadata(methodMetadata);
    }

    /**
     * Analyze the given common metadata.
     * @param commonMetadata the given metadata to analyze.
     * @throws ENCBindingException if the common metadata has not been analyzed
     */
    protected void analyzeCommonMetadata(final ISharedMetadata commonMetadata) throws ENCBindingException {

        // Handle persistence context
        IJavaxPersistenceContext javaxPersistenceContext = commonMetadata.getJavaxPersistenceContext();
        if (javaxPersistenceContext != null) {
            // Set name if null
            if (isNullOrEmpty(javaxPersistenceContext.getName())) {
                javaxPersistenceContext.setName(getName(commonMetadata, javaxPersistenceContext));
            }
            addPersistenceContext(javaxPersistenceContext);
        }

        // Handle persistence unit
        IJavaxPersistenceUnit javaxPersistenceUnit = commonMetadata.getJavaxPersistenceUnit();
        if (javaxPersistenceUnit != null) {
            // Set name if null
            if (isNullOrEmpty(javaxPersistenceUnit.getName())) {
                javaxPersistenceUnit.setName(getName(commonMetadata, javaxPersistenceUnit));
            }
            addPersistenceUnit(javaxPersistenceUnit);
        }

        // Handle resource
        IJAnnotationResource jAnnotationResource = commonMetadata.getJAnnotationResource();
        if (jAnnotationResource != null) {
            // Set name if null
            if (isNullOrEmpty(jAnnotationResource.getName())) {
                jAnnotationResource.setName(getName(commonMetadata, jAnnotationResource));
            }
            jAnnotationResource.setType(getType(commonMetadata, jAnnotationResource.getType()));
            addResource(jAnnotationResource);
        }

        // Handle EJB
        IJEjbEJB jEjbEJB = commonMetadata.getJEjbEJB();
        if (jEjbEJB != null) {
            // Set name if null
            if (isNullOrEmpty(jEjbEJB.getName())) {
                jEjbEJB.setName(getName(commonMetadata, jEjbEJB));
            }
            jEjbEJB.setBeanInterface(getType(commonMetadata, jEjbEJB.getBeanInterface()));
            addEJB(jEjbEJB);
        }


    }

    /**
     * Checks if the given string is empty or null.
     * @param name the given string to test
     * @return true if the string is null or empty
     */
    private boolean isNullOrEmpty(final String name) {
        return (name == null) || "".equals(name);
    }

    /**
     * Gets the ENC name for the given metadata.
     * @param commonMetadata the metadata to analyze.
     * @param o the given object associated to the metadata (used for logging).
     * @return the given ENC name
     * @throws ENCBindingException if no name has been defined on the given name
     *         and the metadata is invalid.
     */
    private String getName(final ISharedMetadata commonMetadata, final Object o) throws ENCBindingException {
        if (commonMetadata instanceof CommonFieldMetadata) {
            CommonFieldMetadata<?, ?, ?> fieldAnnotationMetadata = (CommonFieldMetadata<?, ?, ?>) commonMetadata;
            return fieldAnnotationMetadata.getClassMetadata().getJClass().getName().replace("/", ".") + "/"
                    + fieldAnnotationMetadata.getJField().getName();
        }

        if (commonMetadata instanceof CommonMethodMetadata) {
            CommonMethodMetadata<?, ?, ?> methodAnnotationMetadata = (CommonMethodMetadata<?, ?, ?>) commonMetadata;

            StringBuilder propertyBuilder = new StringBuilder(methodAnnotationMetadata.getJMethod().getName());
            propertyBuilder.delete(0, GETTER_LENGTH);
            propertyBuilder.setCharAt(0, Character.toLowerCase(propertyBuilder.charAt(0)));
            propertyBuilder.insert(0, methodAnnotationMetadata.getClassMetadata().getJClass().getName().replace("/", ".")
                    + "/");
            return propertyBuilder.toString();
        }

        // Last case is the Class case.
        throw new ENCBindingException("There is no defined name in the object '" + o + "' that is defined on the class '"
                + commonMetadata + "'.");

    }

    /**
     * Gets the type for the given metadata.
     * @param commonMetadata the metadata to analyze.
     * @param currentType the actual type of the selected annotation
     * @return the given type value (interface)
     * @throws ENCBindingException if no name has been defined on the given name
     *         and the metadata is invalid.
     */
    private String getType(final ISharedMetadata commonMetadata,
            final String currentType)
            throws ENCBindingException {
        if (commonMetadata instanceof CommonFieldMetadata) {
            CommonFieldMetadata<?, ?, ?> fieldAnnotationMetadata = (CommonFieldMetadata<?, ?, ?>) commonMetadata;
            Type typeInterface = Type.getType(fieldAnnotationMetadata.getJField().getDescriptor());
            return typeInterface.getClassName();
        } else if (commonMetadata instanceof CommonMethodMetadata) {
            CommonMethodMetadata<?, ?, ?> methodAnnotationMetadata = (CommonMethodMetadata<?, ?, ?>) commonMetadata;

            JMethod jMethod = methodAnnotationMetadata.getJMethod();
            Type[] args = Type.getArgumentTypes(jMethod.getDescriptor());
            if (args.length != 1) {
                throw new ENCBindingException("Method args '" + Arrays.asList(args) + "' for method '" + jMethod
                        + "' are invalid. Length should be of 1.");
            }
            Type typeInterface = args[0];
            return typeInterface.getClassName();
        }

        return currentType;
    }

    /**
     * Adds the given persistence context to the list of bindings.
     * @param javaxPersistenceContext the given persistence context.
     */
    protected void addPersistenceContext(final IJavaxPersistenceContext javaxPersistenceContext) {
        this.bindingHolder.addPersistenceContextBinding(new ENCBinding<IJavaxPersistenceContext>(javaxPersistenceContext
                .getName(), javaxPersistenceContext));
    }

    /**
     * Adds the given persistence unit to the list of bindings.
     * @param javaxPersistenceUnit the given persistence unit.
     */
    protected void addPersistenceUnit(final IJavaxPersistenceUnit javaxPersistenceUnit) {
        this.bindingHolder.addPersistenceUnitBinding(new ENCBinding<IJavaxPersistenceUnit>(javaxPersistenceUnit.getName(),
                javaxPersistenceUnit));
    }

    /**
     * Adds the given resource to the list of bindings.
     * @param jAnnotationResource the given resource.
     */
    protected void addResource(final IJAnnotationResource jAnnotationResource) {
        this.bindingHolder.addResourceBinding(new ENCBinding<IJAnnotationResource>(jAnnotationResource.getName(),
                jAnnotationResource));
    }

    /**
     * Adds the given EJB to the list of bindings.
     * @param jEjbEJB the given EJB.
     */
    protected void addEJB(final IJEjbEJB jEjbEJB) {
        this.bindingHolder.addEJBBinding(new ENCBinding<IJEjbEJB>(jEjbEJB.getName(), jEjbEJB));
    }


    /**
     * @return the Holder of all ENC bindings.
     */
    protected ENCBindingHolder getHolder() {
        return this.bindingHolder;
    }

    /**
     * Found all ENC bindings from the given set of metadata.
     * @param warArchiveMetadata the given metadata to analyze.
     * @return object that contains all bindings found for the given set of
     *         classes.
     * @throws ENCBindingException if the analyze of the set of the metadata has
     *         failed.
     */
    public static ENCBindingHolder analyze(final IWarDeployableMetadata warArchiveMetadata) throws ENCBindingException {
        ENCBindingBuilder bindingBuilder = new ENCBindingBuilder(warArchiveMetadata);
        bindingBuilder.analyze();
        return bindingBuilder.getHolder();
    }
}
